/* $Workfile: EnumOtxDevices.c $       $Revision: 11 $      $Modtime: 1/24/03 9:40a $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver Test Program

  @doc PublicDemo

  @module

    ENUMOTXDEVICES.C

    A simple example console application demonstrating the use of OTX Driver
    with the OTX boards.

		The application enumerates board serial number and available
    physical devices.
		
    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"

#define DEMO_DEFAULT_LOG_FILE "EnumOtxDevices.log"
#define DEMO_DEFAULT_BOARD_NO 0

typedef struct DemoEnvT {
  char           *logFileName;
  int             boardNo;
} DemoEnvT;

static int ms_nIndentLevel = 0;

static const char *ms_sqPhysicalDeviceTypeStrings[] = {
  "UNKNOWN",
  "DRIVER",
  "ARNI_PCI",
  "BALDER_PCMCIA",
  "BALDER_S_PCI",
  "BALDER_U_PCI",
  "THOR_PCI",
  "THOR_PCMCIA",
  "VIDAR_5x_PCI",
  "VIDAR_6x_PCI",
  "SIF_PCI",
  "VIDAR_5x_ASM",
  "VIDAR_6x_ASM",
  "VIDAR_55x4_ASM",
  "LI_CONTROLLER",
  "LI_POTS",
  "LI_SLIC",
  "LI_T1E1",
  "LI_BRI_S",
  "LI_BRI_U",
  "HDLC",
  "TSS",
  "DSP",
  "PROCESSOR",
  "PCI_BRIDGE",
  "CODEC",
  "EEPROM",
  "TBC",
  "QUAD_T1E1",
  "QUAD_CODEC",
  "THOR_8_PCI",
  "CT_BUS",
  "HERMOD_PCI",
  "GIMLE_16_PCI",
  "VIDAR_5X1_ASM_BASIC",
  "THOR_1_PCI-BASIC",
  "VIDAR_CST_ASM",      
  "VIDAR_V90_ASM",      
  "VIDAR_510_ASM",      
  "THOR_8_PCI_PLUS",    
  "THOR_2_PCI_PLUS",    
  "SLEIPNIR_1_PCI",     
  "VIDAR-55X12_PCI"    
};


static const char *ms_sqLogicalDeviceTypeStrings[] = {
  "UNKNOWN",
  "EVENT_QUEUE",
  "PIPE",
  "TONE_CONST_VAL_GENERATOR",
  "TONE_EFFECTS_GENERATOR",
  "TONE_SINEWAVE_GENERATOR",
  "TONE_DTMF_GENERATOR",
  "TONE_MF_GENERATOR",
  "TONE_SEQUENCE_GENERATOR",
  "TONE_DTMF_DIALER",
  "TONE_MF_DIALER",
  "TONE_SILENCE_DETECTOR",
  "TONE_DIALTONE_DETECTOR",
  "TONE_DTMF_DETECTOR",
  "TONE_MF_DETECTOR",
  "TONE_FSK_DETECTOR",
  "DATA_RAW_SENDER",
  "DATA_RAW_RECEIVER",
  "DATA_CONVERTER",
  "DATA_MIXER",
  "DATA_CONFERENCE_BRIDGE",
  "HDLC_SENDER",
  "HDLC_RECEIVER",
  "T1_ROB_BIT_ACCESS",
  "E1_S_BIT_ACCESS",
  "E1_ABCD_BIT_ACCESS",
  "MODEM_V34_SENDER",
  "MODEM_V34_RECEIVER",
  "FAX_V17_SENDER",
  "FAX_V17_RECEIVER",
  "VOICE_ECHO_CANCELLER",
  "VOICE_SILENCE_SUPRESSOR",
  "VOICE_CODEC_G723_ENCODER",
  "VOICE_CODEC_G723_DECODER",
  "VOICE_CODEC_G729_ENCODER",
  "VOICE_CODEC_G729_DECODER",
  "USER_APPLICATION",
  "TONE_VAL_SAMPLER",
  "T1_DL_BIT_ACCESS",
  "TONE_DUAL_DETECTOR",
  "TONE_SINGLE_DETECTOR",
};

/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoLogf(char *formatStr, ...);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static OTX_INT32 DemoProcessUserCommand(char commandChar);
static void ReadSerialNum(OTX_HANDLE hParentDevice, OTX_HANDLE hChildDevice, OTX_INT32 eDevType);
static void EnumerateDevices(OTX_HANDLE hParentDevice);
static const char *DemoGetDeviceTypeString(IN OTX_HANDLE eDevType);
static signed int DemoGetDeviceNo(IN OTX_HANDLE hDevice);
static void DemoLogIndent();


/*******************************************************************************
 Global Data
*******************************************************************************/
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;

/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) {
    vfprintf(g_pDemoLogFile,   formatStr, argList);
  }

  va_end(argList);

}


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->logFileName       = DEMO_DEFAULT_LOG_FILE;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;

  for (i = 1; i < argc; i++) {									// Scan command line arguments
    OsaStdStrUpr(argv[i]);
    if (*argv[i] == '-' || *argv[i] == '/') {		// Is argument a legal switch?
      if (*(argv[i]+1) == 'L') {								// Log file
        envP->logFileName = argv[i]+2;
      } else if (*(argv[i]+1) == 'B') {					// Board number
        envP->boardNo = strtoul(argv[i]+2, NULL, 0);
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
				exit(1);
      }
    } else {
      DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
      DemoHelpPanel();
      OsaConsoleExit();
			exit(1);
    }
  }

}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  short c;
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d): %s \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
      } else {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d) Cannot translate this result code \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber);
      }
    }


    DemoLogf("Do you want to exit (Y/N) ");
    c = getch();
    if (c == 'y' || c == 'Y') {
      // Clean up
      OtxDrvReleaseLib();
      OsaConsoleExit();
			exit(1);
    }

  }
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName> : Log file name\n");
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
static OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" a   Test attributes\n");
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  case 'a':
    {
      OtxVersionInfo sVersionInfo;
      char szVersionInfo[255];
      OtxDrvGetAttributeValueSz(0, OTX_ATTR_DRV_REV_STR, szVersionInfo, 255);
      OtxDrvGetAttributeValue(0, OTX_ATTR_DRV_REV_NUM, &sVersionInfo, sizeof(OtxVersionInfo), NULL);

      DemoLogf("Driver version string     %s\n", szVersionInfo);
      DemoLogf("Driver major version      %d\n", sVersionInfo.nMjRev);
      DemoLogf("Driver minor version      %d\n", sVersionInfo.nMnRev);
      DemoLogf("Driver status             %s\n", sVersionInfo.bRelStatus ? "Released" : "Preliminary");
      DemoLogf("Driver preliminary level  %d\n", sVersionInfo.nPrelLevel);
    }

    EnumerateDevices(0);
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}



/*******************************************************************************

  @func

    Read board serial number

*******************************************************************************/
static void ReadSerialNum(OTX_HANDLE hParentDevice, OTX_HANDLE hChildDevice, 
                          OTX_INT32 eDevType)
{
  OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
	OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
	OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
	OTX_RESULT      result;
	char *pszSerialNumber;
				
  if( hParentDevice == 0 && eDevType != -1)		// Parent device with children
	{
    result = OtxDrvGetAttributeType(hChildDevice, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(hChildDevice, result);
		assert(eAttribType == OTX_ATTRIB_SZ);
		assert(bReadOnly == OTX_TRUE);
		pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
		result = OtxDrvGetAttributeValueSz(hChildDevice, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(hChildDevice, result);
		DemoLogf("\nBoard Serial Number: %s\n", pszSerialNumber);
		free(pszSerialNumber);
	}

}


/*******************************************************************************

  @func

  Enumerates all devices for a given parent devices and prints attributes
  etc.

  @comm

*******************************************************************************/
static void EnumerateDevices(OTX_HANDLE hParentDevice)
{
  OTX_UINT32   eDevType;
  OTX_ITERATOR itrChildTypes;
  OTX_ITERATOR itrChildrenOfSameType;

  // This is a recursive function so we indent one more level for each recursion

  // Initialize the opaque iterator for the hosted types
  OtxDrvIteratorBegin(&itrChildTypes);
  while (OTX_SUCCESS(OtxDrvEnumDeviceType(hParentDevice, &itrChildTypes, (OTX_INT32 *)&eDevType))) {   // Note that we must use a zero handle to get the children of the root level
    OTX_HANDLE hChildDevice;
    // Initialize the opaque iterator for the devices within the same type
    OtxDrvIteratorBegin(&itrChildrenOfSameType);
    while (OTX_SUCCESS(OtxDrvEnumDevice(hParentDevice, eDevType, &itrChildrenOfSameType, &hChildDevice))) {
      if (hChildDevice != OTX_INVALID_HANDLE_VALUE) {
        OTX_UINT32 eDevType2;
        OTX_HANDLE hParentDevice2;

				// Here we have a handle for a child device, get the attributes
        DemoLogIndent();
        DemoLogf("%s_%d (%s, handle = %#x)\n", 
          DemoGetDeviceTypeString(hChildDevice), 
          DemoGetDeviceNo(hChildDevice), 
          (eDevType >= 0) ? (eDevType == 0 ? "UnKnown" : "Physical") : "Logical", 
          hChildDevice);

        // Verify that the type and host attributes correspond to the ones we have since earlier
        OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_DEVICE_TYPE, &eDevType2);
        OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_DEVICE_HOST, &hParentDevice2);
        if (eDevType2 != eDevType) {
          DemoLogIndent();DemoLogf("  Incorrect attribute: OTX_ATTR_DEVICE_TYPE\n");
        }
        if (hParentDevice2 != hParentDevice) {
          DemoLogIndent();DemoLogf("  Incorrect attribute: OTX_ATTR_DEVICE_HOST\n");
        }

        // Print device specific attributes
        switch (eDevType) {
        case OTX_LDEVICE_PIPE:
          {
            OTX_UINT32 nValue;
            OTX_UINT8  sqConnMask[OTX_ANY_PIPE_TIMESLOTS_MAX];
            unsigned int i;

            OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_PIPE_CAPACITY, &nValue);
            DemoLogIndent();DemoLogf("  Capacity = %d\n", nValue);

            OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_PIPE_IN_CONNECTED, &nValue);
            if (nValue != OTX_INVALID_HANDLE_VALUE) {
              DemoLogIndent();DemoLogf("  Input connected to device with handle = %#x (%s_%d)\n", nValue, DemoGetDeviceTypeString(nValue), DemoGetDeviceNo(nValue));
              OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_PIPE_IN_HIGHWAY, &nValue);
              DemoLogIndent();DemoLogf("  Input connected to highway = %d\n", nValue);
              OtxDrvGetAttributeValue(hChildDevice, OTX_ATTR_PIPE_IN_CONN_MASK, sqConnMask, OTX_ANY_PIPE_TIMESLOTS_MAX, &nValue);
              DemoLogIndent();DemoLogf("  Input connection mask = ");
              for (i = 0; i < nValue; i++) {
                DemoLogf(" %0.2x", sqConnMask[i]);
              }
              DemoLogf("\n");

            } else {
              DemoLogIndent();DemoLogf("  Input not connected\n");
            }

            OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_PIPE_OUT_CONNECTED, &nValue);
            if (nValue != OTX_INVALID_HANDLE_VALUE) {
              
              DemoLogIndent();DemoLogf("  Output connected to device with handle = %#x (%s_%d)\n", nValue, DemoGetDeviceTypeString(nValue), DemoGetDeviceNo(nValue));
              OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_PIPE_OUT_HIGHWAY, &nValue);
              DemoLogIndent();DemoLogf("  Output connected to highway = %d\n", nValue);

              OtxDrvGetAttributeValue(hChildDevice, OTX_ATTR_PIPE_OUT_CONN_MASK, sqConnMask, OTX_ANY_PIPE_TIMESLOTS_MAX, &nValue);
              DemoLogIndent();DemoLogf("  Output connection mask = ");
              for (i = 0; i < nValue; i++) {
                DemoLogf(" %0.2x", sqConnMask[i]);
              }
              DemoLogf("\n");
            } else {
              DemoLogIndent();DemoLogf("  Output not connected\n");
            }

          }
          break;
        case OTX_LDEVICE_EVENT_QUEUE:
          break;
        case OTX_DEVICE_LI_POTS:
          break;
        case OTX_DEVICE_ARNI_PCI:
        case OTX_DEVICE_BALDER_PCMCIA:
        case OTX_DEVICE_BALDER_S_PCI:
        case OTX_DEVICE_BALDER_U_PCI:
        case OTX_DEVICE_THOR_PCMCIA:
        case OTX_DEVICE_VIDAR_5x_PCI:
        case OTX_DEVICE_VIDAR_6x_PCI:
        case OTX_DEVICE_SIF_PCI:
        case OTX_DEVICE_THOR_PCI:
        case OTX_DEVICE_THOR_8_PCI:
          {
           	OTX_RESULT      result;

            // Read serial number from the board
            ReadSerialNum( hParentDevice, hChildDevice, eDevType );

            // Just to demonstrate the use of OtxDrvOpenPhysicalDeviceEx(). For simplicity we use the DeviceNo as the nEventSourceId, and no associated event queue (OTX_INVALID_HANDLE_VALUE) in this call.
            if (eDevType) {
              DemoLogf("Opening Thor-2-PCI board# %d using OtxDrvOpenPhysicalDeviceEx()\n", DemoGetDeviceNo(hChildDevice));
              
              result = OtxDrvOpenPhysicalDeviceEx(hChildDevice, DemoGetDeviceNo(hChildDevice), OTX_INVALID_HANDLE_VALUE); DemoCheckRc(hChildDevice, result);
            }
          }
          break;
        }

        // Now do the same thing for all children of this device (recursion)
        ms_nIndentLevel += 1;
        EnumerateDevices(hChildDevice);
        ms_nIndentLevel -= 1;
      } else {
        // This should not happen
        DemoLogf("Internal Error: queried a handle value but it still didn't work??? Should not happen!\n");
      }
    }
  }
}


// Returns the string representing the device type for a specific device, automatically chooses between logical and physical devices
static const char *DemoGetDeviceTypeString(IN OTX_HANDLE hDevice)
{
  OTX_INT32 eDevType;

  assert(sizeof(ms_sqPhysicalDeviceTypeStrings) / sizeof(ms_sqPhysicalDeviceTypeStrings[0]) == (size_t) OTX_DEVICE_END);
  assert(sizeof(ms_sqLogicalDeviceTypeStrings) / sizeof(ms_sqLogicalDeviceTypeStrings[0]) == (size_t) OTX_LDEVICE_END);

  if (OTX_SUCCESS(OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, (OTX_UINT32 *)&eDevType))) {
    if (eDevType >= 0) {
      if (eDevType < (OTX_INT32) sizeof(ms_sqPhysicalDeviceTypeStrings)) {
        return ms_sqPhysicalDeviceTypeStrings[eDevType];
      }
    } else if (eDevType < 0) {
      eDevType = -eDevType;
      if (eDevType < (OTX_INT32) sizeof(ms_sqLogicalDeviceTypeStrings)) {
        return ms_sqLogicalDeviceTypeStrings[eDevType];
      }
    }
  }
  return "*** Error ***";
}


// Returns the device no for a specific device
static signed int DemoGetDeviceNo(IN OTX_HANDLE hDevice)
{
  OTX_UINT32 nDevNo;
  if (OTX_SUCCESS(OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDevNo))) {
    return nDevNo;
  } else {
    return -1;
  }
}


/*******************************************************************************

  @func

  Indents according to ms_nIndentLevel, before a DemoLogf()

  @comm

*******************************************************************************/
static void DemoLogIndent()
{
  fprintf(stdout, "%*s", ms_nIndentLevel*2, "");

  if (g_pDemoLogFile) {
    fprintf(g_pDemoLogFile, "%*s", ms_nIndentLevel*2, "");
  }
}



/*******************************************************************************
	main
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

	OsaConsoleInit();

  DemoLogf("EnumOtxDevices, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("OTX Test Program.\n");

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
		OsaConsoleExit();
    exit(-1);
  }

  DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);
  DemoLogf("\nPress 'h' for help\n");

  // Connect to the Driver
  DemoLogf("Connecting lib...");
  result = OtxDrvConnectLib(); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  DemoLogf("Lib connected okay\n");

  while (!exitFlag) {
    commandChar = getch();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}





