/* $Workfile: ThorDemo.c $       $Revision: 105 $      $Modtime: 3/05/03 4:44p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    THORDEMO.C

    A simple example console application demonstrating the use of OTX Driver
    with the Thor-2-PCI, Thor-8-PCI and Thor-2-PCMCIA boards.

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 2 E1 Lines and expects them to be looped with an external
    loop cable. The LI Internal HDLC controller is used to send HDLC frames from
    one Line Interface to be received by the other.

    4 Handsets are connected to the codecs and a through connection is
    demonstrated through the Loop Cable (not available for Thor-2-PCMCIA).

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxDsp.h"

#define DEMO_DEFAULT_LOG_FILE "ThorDemo.log"
#define DEMO_DEFAULT_BOARD_NO 0

#define DEMO_TASK_REF 1L               // Task Reference for a non-blocking call

#define DEMO_SS7_FISU_FILTERING 1

#define DEMO_MAX_MULTIPLE_FRAME_READ 128

//#define DEMO_USE_DSP0_FOR_HDLC0 1  // Uncomment this if you want to use DSP0 (on Thor-2-PCMCIA) has the host for HDLC device 0 (instead of Li0)

typedef struct DemoEnvT {
  char           *logFileName;
  OTX_INT32       nBoardType;
  OtxDataFormatE  law;
  int             boardNo;
  int             coLi;
  OtxT1E1LiModeE  eT1E1Mode;    // T1 or E1 mode
  OTX_BOOL        bSs7FisuFiltering;
  OTX_BOOL        bLoopHandsetsInternally;
  char           *pszDongleType;
  OTX_UINT16      nLineLength;
} DemoEnvT;

/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_BOOL    g_bHdlcVerbose = OTX_TRUE;
OTX_UINT32  g_coHdlcRxFrames[32];
OTX_UINT32  g_coHdlcBadRxFrames[32];
OTX_UINT8   g_sqHdlcFrame[OD_HDLC_MAX_TX_HDLC_FRAME_LEN];
OTX_INT32   g_nSleepTimeBetweenHdlcFrames = 0;
OTX_INT32   g_cbHdlcFrame = 10;
OTX_INT32   g_coHdlcSenderFrames = 1;

struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[2]; // Used for Thor-8-PCI (not for Thor-2-PCI)
  OTX_HANDLE hTss0;
  OTX_HANDLE hLi[8];
  OTX_HANDLE hCodec[4];
#ifdef DEMO_USE_DSP0_FOR_HDLC0
  OTX_HANDLE hDsp[2]; // Used for Thor-2-PCMCIA
#endif
} g_hPhysDevs;


struct {
  OTX_HANDLE hLiCodec[2];
  OTX_HANDLE hCodecLi[2];
  OTX_HANDLE hCodecCodec[2];
  OTX_HANDLE hLiIn[8];
  OTX_HANDLE hLiOut[8];
  OTX_HANDLE hHs3MicPipe;
  OTX_HANDLE hHs3SpkrPipe;
  OTX_HANDLE hHs2MicPipe;
  OTX_HANDLE hHs2SpkrPipe;
} g_hPipes;


struct {
  OTX_HANDLE hHdlcSender[8];
  OTX_HANDLE hHdlcReceiver[8];
} g_hLogicDevs;


struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;
                               
  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (result == OTX_E_NOTIMPL) {
    g_hNotificationEvent = NULL;    // D16V86 apps cannot use event objects  
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the Thor-2-PCI board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1; // Only used for Thor-8-PCI
  OTX_UINT32 nDeviceId;
  OTX_UINT32 nSubsystemId;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
    DemoLogf("Opening Thor-2-PCMCIA board\n");
    break;
  case OTX_DEVICE_THOR_PCI:
    DemoLogf("Opening Thor-2-PCI board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI:
    DemoLogf("Opening Thor-8-PCI board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
    // Get a handle to the Time-Space Switch (TSS) on the Thor-2-PCI board
    DemoLogf("Open OTX_DEVICE_TSS\n");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_TSS, 0, 0, g_hEventQueue, &g_hPhysDevs.hTss0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Open OTX_DEVICE_TSS\n");
    result = OtxDrvEnable(g_hPhysDevs.hTss0, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hTss0, result);
  }

  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_DEVICE_ID, &nDeviceId); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SUBSYSTEM_ID, &nSubsystemId); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  DemoLogf("Board Device ID: %d, SubsystemId: 0x%.4X\n", nDeviceId, nSubsystemId);

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) {
    DemoLogf("Configuring Thor-2-PCMCIA board for dongle type %s\n", g_demoEnv.pszDongleType);
    result = OtxDrvSetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_PCMCIA_DONGLE_TYPE, g_demoEnv.pszDongleType); DemoCheckRc(g_hPhysDevs.hThorController0, result);
#ifdef DEMO_USE_DSP0_FOR_HDLC0
    for (i = 0; i < g_demoEnv.coLi; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      result = OtxDspRunSpm(g_hPhysDevs.hDsp[i], OTX_DSP_SPM_OTXHDLC); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
      result = OtxDrvEnable(g_hPhysDevs.hDsp[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    }
#endif

  }

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_THOR_PCI:
    // On the Thor-2-PCI and Thor-2-PCMCIA boards the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    break;
  case OTX_DEVICE_THOR_8_PCI:
    // On the Thor-8-PCI board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      nQuadE1T1 = i/4;
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  

  if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
    // Get handles to Codec devices #0,#1,#2 and #3
    for (i = 0; i < 4; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_CODEC, i, i, g_hEventQueue, &g_hPhysDevs.hCodec[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
  }

}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

}



/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_HANDLE hHostDevice;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create Hdlc Senders and Receivers logical devices hosted in the physical Line Interface devices
    hHostDevice = g_hPhysDevs.hLi[i];
#ifdef DEMO_USE_DSP0_FOR_HDLC0 // For testing
    if (i == 0) {
      hHostDevice = g_hPhysDevs.hDsp[i];
    }
#endif
    result = OtxDrvCreateLogicalDevice(hHostDevice, OTX_LDEVICE_HDLC_RECEIVER, g_hEventQueue, i, &g_hLogicDevs.hHdlcReceiver[i]); DemoCheckRc(hHostDevice, result);
    result = OtxDrvCreateLogicalDevice(hHostDevice, OTX_LDEVICE_HDLC_SENDER, g_hEventQueue, i, &g_hLogicDevs.hHdlcSender[i]); DemoCheckRc(hHostDevice, result);
  }

}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 ts;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create an incoming and outgoing Pipe representing the external T1/E1
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiIn[i]); DemoCheckRc(g_hPipes.hLiIn[i], result);
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiOut[i]); DemoCheckRc(g_hPipes.hLiOut[i], result);

    // Connect external T1/E1 pipes. Only connect one end of each pipe. External Interface is LI Highway Number 1
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      ts = 24; 
    } else {
      // E1 mode
      ts = 16; 
    }
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[ts] = 0xFF;

#ifdef DEMO_USE_DSP0_FOR_HDLC0
    if (i == 0) {
      result = OtxDrvConnectPipe(g_hPipes.hLiIn[i], OTX_PIPE_INPUT, g_hPhysDevs.hLi[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiIn[i], result);
      result = OtxDrvConnectPipe(g_hPipes.hLiIn[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiIn[i], result);
      result = OtxDrvConnectPipe(g_hPipes.hLiOut[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiOut[i], result);
      result = OtxDrvConnectPipe(g_hPipes.hLiOut[i], OTX_PIPE_INPUT, g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiOut[i], result);
    } else {
      result = OtxDrvConnectPipe(g_hPipes.hLiIn[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiIn[i], result);
      result = OtxDrvConnectPipe(g_hPipes.hLiOut[i], OTX_PIPE_INPUT, g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiOut[i], result);
    }
#else
    result = OtxDrvConnectPipe(g_hPipes.hLiIn[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiIn[i], result);
    result = OtxDrvConnectPipe(g_hPipes.hLiOut[i], OTX_PIPE_INPUT, g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiOut[i], result);
#endif

    // Connect HDLC Sender and Receiver to the external T1/E1 Pipes
    result = OtxDrvConnectLogicalDevice(g_hPipes.hLiIn[i], -1, g_hLogicDevs.hHdlcReceiver[i], -1); DemoCheckRc(g_hPipes.hLiIn[i], result);
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hHdlcSender[i], -1, g_hPipes.hLiOut[i], -1); DemoCheckRc(g_hLogicDevs.hHdlcSender[i], result);

  }

  if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {

    if (g_demoEnv.bLoopHandsetsInternally) {
      // Connect codecs internally (via the time-space switch only)
      for (i = 0; i < 4; i++) {
        result = OtxDrvCreatePipe(64, &g_hPipes.hCodecCodec[i]); DemoCheckRc(g_hPipes.hLiCodec[i], result);
        
        // Create voice pipes for Codec0 <--> Codec1 and Codec2 <--> Codec3
        result = OtxDrvConnectPipe(g_hPipes.hCodecCodec[i], OTX_PIPE_INPUT,  g_hPhysDevs.hCodec[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hCodecCodec[i], result);
        if (i % 2) {
          // Odd Codec
          result = OtxDrvConnectPipe(g_hPipes.hCodecCodec[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[i-1], -1, NULL, 0); DemoCheckRc(g_hPipes.hCodecCodec[i], result);
        } else {
          // Even Codec
          result = OtxDrvConnectPipe(g_hPipes.hCodecCodec[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[i+1], -1, NULL, 0); DemoCheckRc(g_hPipes.hCodecCodec[i], result);
        }
        
        
      }
    } else { 
      // Connect codecs via the T1/E1 span
      
      for (i = 0; i < 4; i++) {
        // Create voice pipes for Li0 <--> Codec0 and Li1 <--> Codec1 and Li0 <--> Codec2 and Li1 <--> Codec3
        result = OtxDrvCreatePipe(64, &g_hPipes.hLiCodec[i]); DemoCheckRc(g_hPipes.hLiCodec[i], result);
        result = OtxDrvCreatePipe(64, &g_hPipes.hCodecLi[i]); DemoCheckRc(g_hPipes.hCodecLi[i], result);
        
        // Connect Li to Codec, full duplex. Don't care Highway number for T1/E1 Li reverts to internal highway
        // Codec0 and Codec1 --> timeslot 10 and Codec2 and Codec3 --> timeslot 11
        ts = 10 + i/2;
        memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
        sqTimeSlotMask[ts] = 0xFF;
        
        result = OtxDrvConnectPipe(g_hPipes.hLiCodec[i], OTX_PIPE_INPUT,  g_hPhysDevs.hLi[i%2], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiCodec[i], result);
        result = OtxDrvConnectPipe(g_hPipes.hLiCodec[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hLiCodec[i], result);
        
        result = OtxDrvConnectPipe(g_hPipes.hCodecLi[i], OTX_PIPE_INPUT,  g_hPhysDevs.hCodec[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hCodecLi[i], result);
        result = OtxDrvConnectPipe(g_hPipes.hCodecLi[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i%2], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hCodecLi[i], result);
        
      }
    }
  }


}




/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Initialize the Hdlc Sender and Receiver
    result = OtxHdlcSenderInit(
               g_hLogicDevs.hHdlcSender[i], // Handle to logical HDLC Filter Device
               OTX_HDLC_FFT_FLAGS,          // Fill character(s) to be send between HDLC frames
               2,                           // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
               OTX_HDLC_CRC16,              // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
               OTX_FALSE,                   // If OTX_TRUE then all bits are inverted when transmitted and before received
               OTX_TRUE,                    // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
               OTX_HDLC_TR_NA);             // HDLC Transmission Rate for V.110 and X.30 modes
               DemoCheckRc(g_hLogicDevs.hHdlcSender[i], result);

    result = OtxHdlcReceiverInit(
               g_hLogicDevs.hHdlcReceiver[i], // Handle to logical HDLC Filter Device
               OTX_HDLC_FFT_FLAGS,          // Fill character(s) to be send between HDLC frames
               2,                           // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
               OTX_HDLC_CRC16,              // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
               OTX_FALSE,                   // If OTX_TRUE then all bits are inverted when transmitted and before received
               OTX_TRUE,                    // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
               OTX_HDLC_TR_NA,              // HDLC Transmission Rate for V.110 and X.30 modes
               OTX_FALSE);                   // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
               DemoCheckRc(g_hLogicDevs.hHdlcReceiver[i], result);

    if (g_demoEnv.bSs7FisuFiltering) {
      result = OtxHdlcSS7SetFilter(g_hLogicDevs.hHdlcReceiver[i], OTX_HDLC_SS7_FILTER_FISU); DemoCheckRc(g_hLogicDevs.hHdlcReceiver[i], result);
    }
  }

}


/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1;

  // Enable Board
  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // If board type is Thor-8-PCI then enable Quad E1T1 device
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) {
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
#if 0 // Enable only Li1
    if (i == 1) {
      continue;
    }
#endif
    result = OtxDrvEnable(g_hPhysDevs.hLi[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }


}


/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
void DemoEnableLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Enable Logical HDLC Devices (Data Sinks)
  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hLogicDevs.hHdlcSender[i],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hHdlcSender[i], result);
    result = OtxDrvEnable(g_hLogicDevs.hHdlcReceiver[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hHdlcReceiver[i], result);
  }
}

/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 128;
  pE1CfgData->nLosRecovery               = 128;
  pE1CfgData->nLineLength                = g_demoEnv.nLineLength;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;    // CCS, CAS_BR;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_AMI;   // AMI, B8ZS
  pT1CfgData->eRxLineCode                = OTX_T1_LC_AMI;   // AMI, B8ZS
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;    // F12 (D3/4 or SF), F4, ESF, F72 (SLC96)

  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_TRUE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_TRUE;
  pT1CfgData->nLosSensitivity            = 128;
  pT1CfgData->nLosRecovery               = 128;
  pT1CfgData->nLineLength                = g_demoEnv.nLineLength;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
#if 1
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
#else
  pT1CfgData->bReceiveEqualizer          = OTX_FALSE;
#endif
  pT1CfgData->bfClearChannels            = 0x000000;
}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  short c;
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d): %s \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
      } else {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d) Cannot translate this result code \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber);
      }
    }


    DemoLogf("Do you want to exit (Y/N) ");
    c = getch();
    if (c == 'y' || c == 'Y') {
      // Clean up
			OsaConsoleExit();
      OtxDrvReleaseLib();
      exit(1);
    }

  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct tm *ptCurrentDayAndTime;
  time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 1164447360 * 100000000;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with localtime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = localtime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}


/*******************************************************************************
  @func

*******************************************************************************/
void  *DemoTimeStampTest(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_DATETIME nTimeStamp;
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct tm *ptCurrentDayAndTime;
  time_t nCurrentTimeSecs;
  
  for (;;) {
    
    result = OtxDrvGetCurrentTime(&nTimeStamp); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    //DemoLogf("%.8X%.8X (Current time)\n", nTimeStamp.sUint32.nHigh, nTimeStamp.sUint32.nLow);
    
    // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
#if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 1164447360 * 100000000;  
#elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
#endif
    
    // Convert to seconds since Jan 1, 1970 (so we can use this time with localtime())
    nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
    
    // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
    nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
    
    nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
    ptCurrentDayAndTime = localtime(&nCurrentTimeSecs);
    
    if (ptCurrentDayAndTime) {
      
      DemoLogf("Current time is (%.2d:%.2d:%.2d.%.3d) %s", 
        ptCurrentDayAndTime->tm_hour,
        ptCurrentDayAndTime->tm_min, 
        ptCurrentDayAndTime->tm_sec,
        nTimeStampMs.sUint32.nLow,
        asctime(ptCurrentDayAndTime)
        );
    } else {
      DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
    }

    OsaThreadSleep(4000);
  }
  


  return 0;

}




/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");

    do {
#if 0 // Single reads of events
      OtxEventDataS eventData;
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (result == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (result == OTX_S_OK);

#else // Multiple reads of events
#define DEMO_EVENT_DATA_BUF_SIZE 10
      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;

      result = OtxDrvGetEventDataEx(g_hEventQueue, sqEventData, DEMO_EVENT_DATA_BUF_SIZE, &coEventsFetched); DemoCheckRc(g_hEventQueue, result);

      if (result == OTX_S_OK) {
        // We have a driver event (one at least)
        //DemoLogf("OtxDrvGetEventDataEx() fetched %d events\n", coEventsFetched);

        assert(coEventsFetched <= coEventsFetched);

        for (i = 0; i < coEventsFetched; i++) {
          // Print the event data.
          // In a real application branch based on the data
          DemoPrintEvent(&sqEventData[i]);
        }

      }
#endif
    } while (result == OTX_S_OK);

  }

  return 0;

}



/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  OTX_UINT32 i;
  char msgStr[256];

  switch (pEventData->m_eDeviceType) {
  case OTX_LDEVICE_HDLC_RECEIVER:
    if (g_bHdlcVerbose) {
      DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    }
    break;
  default:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    break;
  }
    
  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_TSS:
    // Event from the Time-Space Switch
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("TSS Event: %s \n", msgStr);
    } else {
      DemoLogf("TSS Invalid Event: %ld \n", pEventData->m_nCode);
    }
    break;
    
  case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (!strncmp(msgStr, "Invalid", strlen("Invalid"))) {
      DemoLogf("E1/T1[%d] Code is %d\n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

#if 0 // Single Frame Read
  case OTX_LDEVICE_HDLC_RECEIVER:
    // Event from HLDC Receiver 
    if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
      OTX_UINT8  buf[1024];
      OTX_UINT32 nFrameLen;
      OTX_UINT32 nFrameSeq;
      OTX_UINT32 nFrameEndCode;
      OTX_DATETIME nTimeStamp;
      
      
      otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
      while (otxRc == OTX_S_OK) {
              
        if (pEventData->m_nSourceId < 32) {
          g_coHdlcRxFrames[pEventData->m_nSourceId]++;
        }
        
        
        if (g_bHdlcVerbose) {
          DemoLogf("HDLC Frame Received, Sequence Number = %d\n", pEventData->m_nParam);
          
          DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
          for (i = 0; i < nFrameLen && i < 1024; i++) {
            DemoLogf("%.2X ", buf[i]);
          }
          DemoLogf("\n");
        } else {
          if (nFrameEndCode) {
            
            g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
            
            DemoLogf("FrameEndCode Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
            for (i = 0; i < nFrameLen && i < 256; i++) {
              DemoLogf("%.2X ", buf[i]);
            }
            DemoLogf("\n");
            
            if (g_coHdlcBadRxFrames[pEventData->m_nSourceId] > 100) {
              DemoLogf("Too many bad frames from HdlcReceiver[%d]. Disabling device\n", pEventData->m_nSourceId);
              otxRc = OtxDrvDisable(pEventData->m_hDevice,  OTX_TASK_SYNC, 5000L); DemoCheckRc(pEventData->m_hDevice, otxRc);
            }
            
            
          } else {
            if ((pFrame->cbOctets != (OTX_UINT32)g_cbHdlcFrame) && (pFrame->cbOctets != 3)) {
              g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
              
              DemoLogf("LENGTH Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
              for (i = 0; i < nFrameLen && i < 256; i++) {
                DemoLogf("%.2X ", buf[i]);
              }
              DemoLogf("\n");
              
            } else if (memcmp(pBuf+3, g_sqHdlcFrame+3, g_cbHdlcFrame-3)) {
              g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
              
              DemoLogf("memcmp Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
              for (i = 0; i < nFrameLen && i < 256; i++) {
                DemoLogf("%.2X ", buf[i]);
              }
              DemoLogf("\n");
              
            }
          }
        } 
        
        // Read next frame (if any)
        otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);

      } 
      if (otxRc != OTX_S_FALSE) {
        DemoLogf("HdlcReceiver[%d] OtxHdlcReadData() was unsuccessful.\n", pEventData->m_nSourceId);
        DemoCheckRc(pEventData->m_hDevice, otxRc);
      }
      
      
    } else {
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
#else // Multiple frame read
  case OTX_LDEVICE_HDLC_RECEIVER:
    // Event from HLDC Receiver 
    if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
#define CB_BUF_SIZE 8192
      OTX_UINT8  buf[CB_BUF_SIZE];
      OTX_UINT8  *pBuf;
			OtxHdlcFrameInfoS *pFrame;
      OTX_UINT32 coFrames;
      OTX_UINT32 j;
      
      otxRc = OtxHdlcReadDataEx(pEventData->m_hDevice, buf, CB_BUF_SIZE, DEMO_MAX_MULTIPLE_FRAME_READ, OTX_FALSE, &coFrames);
      while (otxRc == OTX_S_OK) {
      
        if (pEventData->m_nSourceId < 32) {
          g_coHdlcRxFrames[pEventData->m_nSourceId] += coFrames;
        }

        if (g_bHdlcVerbose) {
          DemoLogf("%d HDLC Frame(s) Received, Sequence Number = %d\n", coFrames, pEventData->m_nParam);
        }
        
        pFrame = (OtxHdlcFrameInfoS *) buf;

        for (j = 0; j < coFrames; j++) {

					pBuf = pFrame->sqData;

          if (g_bHdlcVerbose) {
            DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->nFrameSeq, pFrame->nFrameEndCode, pFrame->nFrameSeq);
            for (i = 0; i < pFrame->cbOctets && i < CB_BUF_SIZE; i++) {
              DemoLogf("%.2X ", pBuf[i]);
            }
            DemoLogf("\n");
          } else {
            if (pFrame->nFrameEndCode) {
              
              g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
              
              DemoLogf("FrameEndCode Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->cbOctets, pFrame->nFrameEndCode, pFrame->nFrameSeq);
              for (i = 0; i < pFrame->cbOctets && i < 256; i++) {
                DemoLogf("%.2X ", pBuf[i]);
              }
              DemoLogf("\n");
              
              if (g_coHdlcBadRxFrames[pEventData->m_nSourceId] > 100) {
                DemoLogf("Too many bad frames from HdlcReceiver[%d]. Disabling device\n", pEventData->m_nSourceId);
                otxRc = OtxDrvDisable(pEventData->m_hDevice,  OTX_TASK_SYNC, 5000L); DemoCheckRc(pEventData->m_hDevice, otxRc);
              }
              
              
            } else {
              if ((pFrame->cbOctets != (OTX_UINT32)g_cbHdlcFrame) && (pFrame->cbOctets != 3)) {
                g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
                
                DemoLogf("LENGTH Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->cbOctets, pFrame->nFrameEndCode, pFrame->nFrameSeq);
                for (i = 0; i < pFrame->cbOctets && i < 256; i++) {
                  DemoLogf("%.2X ", pBuf[i]);
                }
                DemoLogf("\n");
                
              } else if (memcmp(pBuf+3, g_sqHdlcFrame+3, g_cbHdlcFrame-3)) {
                g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
                
                DemoLogf("memcmp Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->cbOctets, pFrame->nFrameEndCode, pFrame->nFrameSeq);
                for (i = 0; i < pFrame->cbOctets && i < 256; i++) {
                  DemoLogf("%.2X ", pBuf[i]);
                }
                DemoLogf("\n");
                
              }
            }
          }  

          // Step pointer to next frame (if any)
					pFrame = OTX_STRUCT_ADD_BYTE_OFFSET(OtxHdlcFrameInfoS, pFrame, pFrame->nOffsetNext);
        }

        // Read next frame (if any)
        otxRc = OtxHdlcReadDataEx(pEventData->m_hDevice, buf, CB_BUF_SIZE, DEMO_MAX_MULTIPLE_FRAME_READ, OTX_FALSE, &coFrames);

      } 
      if (otxRc != OTX_S_FALSE) {
        DemoLogf("HdlcReceiver[%d] OtxHdlcReadDataEx() was unsuccessful.\n", pEventData->m_nSourceId);
        DemoCheckRc(pEventData->m_hDevice, otxRc);
      }
      
      
    } else {
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
#endif

    
  case OTX_LDEVICE_HDLC_SENDER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    DemoLogf("HdlcSender[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    break;
    
  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}

/*******************************************************************************

  @func

*******************************************************************************/
void DemoCloseDevices(
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;

  // Hdlc Senders and Receivers logical devices hosted in the physical Line Interface devices
  for (i = 0; i < g_demoEnv.coLi; i++) {
    if (g_hLogicDevs.hHdlcSender[i]) {
      result = OtxDrvClose(g_hLogicDevs.hHdlcSender[i]); DemoCheckRc(g_hLogicDevs.hHdlcSender[i], result);
    }
    if (g_hLogicDevs.hHdlcReceiver[i]) {
      result = OtxDrvClose(g_hLogicDevs.hHdlcReceiver[i]); DemoCheckRc(g_hLogicDevs.hHdlcReceiver[i], result);
    }
  }

  // Incoming and outgoing Pipe representing the external T1/E1
  for (i = 0; i < g_demoEnv.coLi; i++) {
    if (g_hPipes.hLiIn[i]) {
      result = OtxDrvClose(g_hPipes.hLiIn[i]); DemoCheckRc(g_hPipes.hLiIn[i], result);
    }
    if (g_hPipes.hLiOut[i]) {
      result = OtxDrvClose(g_hPipes.hLiOut[i]); DemoCheckRc(g_hPipes.hLiOut[i], result);
    }
  }

  // Voice pipes 
  for (i = 0; i < 2; i++) {
    if (g_hPipes.hLiCodec[i]) {
      result = OtxDrvClose(g_hPipes.hLiCodec[i]); DemoCheckRc(g_hPipes.hLiCodec[i], result);
    }
    if (g_hPipes.hCodecLi[i]) {
      result = OtxDrvClose(g_hPipes.hCodecLi[i]); DemoCheckRc(g_hPipes.hCodecLi[i], result);
    }
    
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_INT32      nCurrentLi = 0;
  OTX_RESULT result;



  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-1 Select active Li, this line is then used for subsequent commands\n");
    DemoLogf(" t   Send a HDLC fram on the current Li\n");
    DemoLogf(" e/E Enable/Disable current Li\n");
    DemoLogf(" c   Set Clock Source\n");
    DemoLogf(" s   Display the SS7 HDLC receive status of the active Li\n");
    DemoLogf(" l   Query Line Status of both Lis\n");
    DemoLogf(" a/A Start/Stop Sending AIS (All 1's) on current Li\n");
    DemoLogf(" f   Force Resynchronization on both Lis\n");
    DemoLogf(" b   Reset Current HDLC Receiver\n");
    DemoLogf(" B   Reset Current E1T1 device\n");
    DemoLogf(" z   Set T1E1 Line Termination\n");
    DemoLogf(" *   Set Line Build Out (T1 mode only)\n");
    DemoLogf(" &   Read receive equalizer status\n");
    DemoLogf(" ^   Set Receive Input Threshold for current Li\n");
    DemoLogf(" p   Set Loopback mode\n");
    DemoLogf(" o   Buffer 10 HDLC frames to be transmitted, then Flush the HDLC Sender\n");
    DemoLogf(" g   Get current time\n");
    DemoLogf(" m/M Enable/Disable Monitor mode (Thor-8-PCI only)\n");
    DemoLogf(" L   Set HDLC Frame length to be transmitted\n");
    DemoLogf(" +   Set number of consecutive frame to be send from the HDLC Sender\n");
    DemoLogf(" S   Set time to sleep (in ms) between transmitted HDLC Frames (%d)\n", g_nSleepTimeBetweenHdlcFrames);
    DemoLogf(" !   Customize transmit pulse shape for current Li\n");
    DemoLogf(" @   Set idle pattern to 0x01 (to facilitate pulse shape testing)\n"); 
    DemoLogf(" k/K Turn Amp On/Off for current dongle (HMA-1079-2-1.0 dongle only)\n"); 
    //DemoLogf("OTX_RES_END = %d \n", OTX_RES_END);
    //DemoLogf("OTX_T1E1_RES_BAD_RX_FRAME_FORMAT = %d \n", OTX_T1E1_RES_BAD_RX_FRAME_FORMAT);
    break;

  case 'y':
    {

      DemoLogf("Change Li mode of both Lis \n");

      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
        DemoLogf("Switching to E1 mode \n");
        g_demoEnv.eT1E1Mode = OTX_T1E1_MODE_E1;
      } else {
        DemoLogf("Switching to T1 mode \n");
        g_demoEnv.eT1E1Mode = OTX_T1E1_MODE_T1;
      }
      
      DemoCloseDevices();

      DemoConfigureE1T1Devices();
      
      DemoEnablePhysicalDevices();

      DemoCreateLogicalDevices();
      
      DemoConnectDevices();
      
      DemoInitializeLogicalDevices();
      
      DemoEnableLogicalDevices();
      
      DemoLogf("Done.\n");
    }
    break;

  case 'g':
    {
      OTX_DATETIME nTimeStamp;
      OTX_DATETIME nTimeStampMs;
      OTX_DATETIME nTimeStampSec;
      struct tm *ptCurrentDayAndTime;
      time_t nCurrentTimeSecs;
      
      result = OtxDrvGetCurrentTime(&nTimeStamp); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
      
      //DemoLogf("%.8X%.8X (Current time)\n", nTimeStamp.sUint32.nHigh, nTimeStamp.sUint32.nLow);
      
      // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
#if defined(__linux__)
      nTimeStamp.nUint64 = nTimeStamp.nUint64 - 1164447360 * 100000000;  
#elif defined(WIN32)
      nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
#endif
      
      // Convert to seconds since Jan 1, 1970 (so we can use this time with localtime())
      nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
      
      // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
      nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
      
      nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
      ptCurrentDayAndTime = localtime(&nCurrentTimeSecs);
      
      if (ptCurrentDayAndTime) {
        
        DemoLogf("Current time is (%.2d:%.2d:%.2d.%.3d) %s", 
          ptCurrentDayAndTime->tm_hour,
          ptCurrentDayAndTime->tm_min, 
          ptCurrentDayAndTime->tm_sec,
          nTimeStampMs.sUint32.nLow,
          asctime(ptCurrentDayAndTime)
          );
      } else {
        DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
      }
      

    }
    break;

  case 'G':
    {
      OsaThreadStart(DemoTimeStampTest);
    }
    break;


  case 'p':
    {
      OtxT1E1LoopE eLoopType;
      OTX_INT32 nLi;

      DemoLogf("Set Loopback mode for both Line Interfaces \n");

      DemoLogf("0: All Line Loops off                     \n");
      DemoLogf("1: Line Loop                              \n");
      DemoLogf("2: Remote Loop with Jitter Attenuation On \n");
      DemoLogf("3: Remote Loop with Jitter Attenuation Off\n");
      DemoLogf("4: Local Loop                             \n");
      
      switch (getch() - '0') {
      case 0:
        eLoopType = OTX_T1E1_LOOPS_OFF;
        break;
      case 1:
        eLoopType = OTX_T1E1_LINE_LOOP;
        break;
      case 2:
        eLoopType = OTX_T1E1_REMOTE_ON_LOOP_JITTER_ATTEN;
        break;
      case 3:
        eLoopType = OTX_T1E1_REMOTE_OFF_LOOP_JITTER_ATTEN;
        break;
      case 4:
        eLoopType = OTX_T1E1_LOCAL_LOOP;
        break;
      default:
        DemoLogf("Invalid entry. Turning Loop off (OTX_T1E1_LOOPS_OFF)\n");
        eLoopType = OTX_T1E1_LOOPS_OFF;
        break;
      }

      for (nLi = 0; nLi < 2; nLi++) {
        DemoLogf("Setting Loopback mode (%d) on Li #%d \n", (int)eLoopType, nLi);
        result = OtxT1E1SetLineLoop(g_hPhysDevs.hLi[nLi], eLoopType); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
      }

      DemoLogf("Done.");
    }
    break;

  case ':':
    {
      OTX_INT32 i;

      /*
      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
        DemoLogf("Switching both Lis to OTX_T1E1_MODE_E1 mode\n");
        g_demoEnv.eT1E1Mode = OTX_T1E1_MODE_E1;
      } else {
        DemoLogf("Switching both Lis to OTX_T1E1_MODE_T1 mode\n");
        g_demoEnv.eT1E1Mode = OTX_T1E1_MODE_T1;
      }
      */

      for (i = 0; i < 2; i++) {
        
        // Set termination for the T1/E1 Front End
        result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);
        
        // Configure Transceivers
        if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
          // T1 mode
          result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
        } else {
          // E1 mode
          result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
        }
        result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
        
      }

      DemoLogf("Done\n");
  
    }
    break;


  
  case 'f':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < 2; nLi++) {
        DemoLogf("Forcing resynchronization on Li #%d... ", nLi);
        result = OtxT1E1ForceResynch(g_hPhysDevs.hLi[nLi]); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("Done.");
      }
    }
    break;

#if 0 // For debugging
  case 'x':
    {
      OTX_UINT8 sqTimeSlotMask[32];
      OTX_INT32 ts;
      
      DemoLogf("Connecting HS3 to ASM Hwy0, TS0\n");
      ts = 4;
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs3MicPipe, OTX_PIPE_INPUT,   g_hPhysDevs.hTss0, 14, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hHs3MicPipe, result);

      ts = 0;
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs3MicPipe, OTX_PIPE_OUTPUT,   g_hPhysDevs.hTss0, 10, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hHs3MicPipe, result);
      
      DemoLogf("Done.\n");
    }
    break;
#endif





  case 'a':
    {

      DemoLogf("Send AIS (All Ones) on Li #%d \n", nCurrentLi);
      result = OtxT1E1AlarmOn(g_hPhysDevs.hLi[nCurrentLi], OTX_T1E1_ALARM_AIS); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      DemoLogf("Done.\n");
    }
    break;

  case 'A':
    {

      DemoLogf("Stop Sending AIS (All Ones) on Li #%d \n", nCurrentLi);
      result = OtxT1E1AlarmOff(g_hPhysDevs.hLi[nCurrentLi], OTX_T1E1_ALARM_AIS); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      DemoLogf("Done.\n");
    }
    break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Line status of Li #%d is ", nLi);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
    }
    break;

  case 'b':
    {
      DemoLogf("Reseting HdlcReceiver[%d]...\n",nCurrentLi);
      result = OtxDrvReset(g_hLogicDevs.hHdlcReceiver[nCurrentLi]); DemoCheckRc(g_hLogicDevs.hHdlcReceiver[nCurrentLi], result);
      DemoLogf("Done\n");
    }
    break;

  case 'B':
    {
      DemoLogf("Reseting Li #%d \n", nCurrentLi);
      result = OtxDrvReset(g_hPhysDevs.hLi[nCurrentLi]); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      DemoLogf("Done\n");
    }
    break;

  case 's':
    {
      OTX_UINT32  nFisuCount;   // Number of FISUs received
      OTX_UINT32  nLssuCount;   // Number of LSSUs received
      OTX_UINT32  nMsuCount;    // Number of MSUs received
      OtxHdlcSS7FisuS fisu;     // The last received FISU on the pipe.

      result = OtxHdlcSS7GetReceiveStatus(g_hLogicDevs.hHdlcReceiver[nCurrentLi], &nFisuCount, &nLssuCount, &nMsuCount, &fisu); DemoCheckRc(g_hLogicDevs.hHdlcReceiver[nCurrentLi], result);

      if (OTX_SUCCESS(result)) {
		DemoLogf("HdlcReceiver[%d]: SS7 receive statistics\n",nCurrentLi);
        DemoLogf("FISUs received: %lu\n", nFisuCount);
        DemoLogf("LSSUs received: %lu\n", nLssuCount);
        DemoLogf("MSUs  received: %lu\n", nMsuCount);
        DemoLogf("Last FISU received: bsn = %d, bib = %d, fsn = %d, fib = %d, li = %d\n",
               fisu.bsn, fisu.bib, fisu.fsn, fisu.fib, fisu.li);
      }

    }
    break;

#if 0
  case 'm':
    {
      DemoLogf("Disconnecting HS2 Pipes (Spkr and Mic)\n");
      result = OtxDrvDisconnectPipe(g_hPipes.hHs2MicPipe, OTX_PIPE_INPUT); DemoCheckRc(g_hPipes.hHs2MicPipe, result);
      result = OtxDrvDisconnectPipe(g_hPipes.hHs2MicPipe, OTX_PIPE_OUTPUT); DemoCheckRc(g_hPipes.hHs2MicPipe, result);
      result = OtxDrvDisconnectPipe(g_hPipes.hHs2SpkrPipe, OTX_PIPE_INPUT); DemoCheckRc(g_hPipes.hHs2SpkrPipe, result);
      result = OtxDrvDisconnectPipe(g_hPipes.hHs2SpkrPipe, OTX_PIPE_OUTPUT); DemoCheckRc(g_hPipes.hHs2SpkrPipe, result);
      
      DemoLogf("Done.\n");
    }
    break;

  case 'M':
    {
      DemoLogf("Disconnecting HS3 Pipes (Spkr and Mic)\n");
      result = OtxDrvDisconnectPipe(g_hPipes.hHs3MicPipe, OTX_PIPE_INPUT); DemoCheckRc(g_hPipes.hHs3MicPipe, result);
      result = OtxDrvDisconnectPipe(g_hPipes.hHs3MicPipe, OTX_PIPE_OUTPUT); DemoCheckRc(g_hPipes.hHs3MicPipe, result);
      result = OtxDrvDisconnectPipe(g_hPipes.hHs3SpkrPipe, OTX_PIPE_INPUT); DemoCheckRc(g_hPipes.hHs3SpkrPipe, result);
      result = OtxDrvDisconnectPipe(g_hPipes.hHs3SpkrPipe, OTX_PIPE_OUTPUT); DemoCheckRc(g_hPipes.hHs3SpkrPipe, result);
      
      DemoLogf("Done.\n");
    }
    break;

  case 'V':
    {
      OTX_UINT8 sqTimeSlotMask[32];
      OTX_INT32 ts;
      
      DemoLogf("Connecting HS3 to H100 Hwy0 ");
      ts = 4;
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs3MicPipe, OTX_PIPE_INPUT,   g_hPhysDevs.hTss0, 14, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hHs3MicPipe, result);

      if (g_demoEnv.boardNo == 0) {
        ts = 0;
      } else {
        ts = 1;
      }
      DemoLogf("Mic to TS%d, ", ts);
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs3MicPipe, OTX_PIPE_OUTPUT,   g_hPhysDevs.hTss0, 16, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hHs3MicPipe, result);
      
      if (g_demoEnv.boardNo == 0) {
        ts = 1;
      } else {
        ts = 0;
      }
      DemoLogf("Spkr to TS%d \n", ts);
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs3SpkrPipe, OTX_PIPE_INPUT,   g_hPhysDevs.hTss0, 16, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hHs3SpkrPipe, result);

      ts = 4;
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs3SpkrPipe, OTX_PIPE_OUTPUT,   g_hPhysDevs.hTss0, 14, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hHs3SpkrPipe, result);

      DemoLogf("Done.\n");
    }
    break;
#endif

#if 0
  case 'v':
    {
      OTX_UINT8 sqTimeSlotMask[32];
      OTX_INT32 ts;
      
      DemoLogf("Connecting HS2 to H100 Hwy0 ");
      ts = 3;
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs2MicPipe, OTX_PIPE_INPUT,   g_hPhysDevs.hTss0, 14, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hHs2MicPipe, result);

      if (g_demoEnv.boardNo == 0) {
        ts = 0;
      } else {
        ts = 1;
      }
      DemoLogf("Mic to TS%d, ", ts);
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs2MicPipe, OTX_PIPE_OUTPUT,   g_hPhysDevs.hTss0, 16, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hHs2MicPipe, result);
      
      if (g_demoEnv.boardNo == 0) {
        ts = 1;
      } else {
        ts = 0;
      }
      DemoLogf("Spkr to TS%d \n", ts);
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs2SpkrPipe, OTX_PIPE_INPUT,   g_hPhysDevs.hTss0, 16, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hHs2SpkrPipe, result);

      ts = 3;
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs2SpkrPipe, OTX_PIPE_OUTPUT,   g_hPhysDevs.hTss0, 14, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hHs2SpkrPipe, result);

      DemoLogf("Done.\n");
    }
    break;
#endif

  case 'm':
    DemoLogf("Enabling Monitor (-20dB) mode for Li #%d \n", nCurrentLi);
    result = OtxT1E1SetMonitorMode(g_hPhysDevs.hLi[nCurrentLi], OTX_TRUE); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
    DemoLogf("Done.\n");
    break;

  case 'M':
    DemoLogf("Disabling Monitor (-20dB) mode for Li #%d \n", nCurrentLi);
    result = OtxT1E1SetMonitorMode(g_hPhysDevs.hLi[nCurrentLi], OTX_FALSE); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
    DemoLogf("Done.\n");
    break;

  case 'e':
    DemoLogf("Enable Li #%d \n", nCurrentLi);
    result = OtxDrvEnable(g_hPhysDevs.hLi[nCurrentLi], DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
    DemoLogf("Done.\n");
    break;

  case 'E':
    DemoLogf("Disable Li #%d \n", nCurrentLi);
    result = OtxDrvDisable(g_hPhysDevs.hLi[nCurrentLi], DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
    DemoLogf("Done.\n");
    break;

  case 'z':
    {
      OtxT1E1TermModeE eTermMode;

      DemoLogf("Set Line Termination mode:\n");
      DemoLogf("0 = OTX_T1E1_TERM_120_OHM\n");
      DemoLogf("1 = OTX_T1E1_TERM_100_OHM\n");
      DemoLogf("2 = OTX_T1E1_TERM_75_OHM\n");
      DemoLogf("3 = OTX_T1E1_TERM_HIGHZ\n");
      switch (getch() - '0') {
      case 0:
        eTermMode = OTX_T1E1_TERM_120_OHM;
        break;
      case 1:
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      case 2:
        eTermMode = OTX_T1E1_TERM_75_OHM;
        break;
      case 3:
        eTermMode = OTX_T1E1_TERM_HIGHZ;
        break;
      default:
        DemoLogf("Invalid entry. Setting Line Termination to OTX_T1E1_TERM_100_OHM\n");
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      }

      result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[nCurrentLi], eTermMode); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Done.\n");
    }
    break;

  case 't':
    {
      OTX_INT32 i;
      OTX_INT32 busyCount = 0;
      clock_t startTime, stopTime;

      if (g_cbHdlcFrame > OD_HDLC_MAX_TX_HDLC_FRAME_LEN) {
        DemoLogf("Error: Frame Length too long\n");
        break;
      }

      for (i = 0; i < g_cbHdlcFrame; i++) {
        g_sqHdlcFrame[i] = (OTX_UINT8) i;
      }

      DemoLogf("Sending %d HDLC frame(s) from Li #%d (len=%d) ...", g_coHdlcSenderFrames, nCurrentLi, g_cbHdlcFrame);

      // Send HDLC Frame(s)
      startTime = clock();
      for (i = 0; i < g_coHdlcSenderFrames; i++) {
        result = OtxHdlcSendData(g_hLogicDevs.hHdlcSender[nCurrentLi], g_sqHdlcFrame, g_cbHdlcFrame, DEMO_TASK_REF, 200); 
        if (result == OTX_E_BUSY) {
          // Driver IO Request buffer is full. We need to retransmit this frame.
          i--;
          busyCount++;
        } else {
          DemoCheckRc(g_hLogicDevs.hHdlcSender[nCurrentLi], result);
          OsaThreadSleep(g_nSleepTimeBetweenHdlcFrames);
        }
      }
      stopTime = clock();

      DemoLogf("Done (BusyCount=%d, time=%2.1f seconds).\n", busyCount, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
    }
    break;


  case 'o':
    {
      OTX_UINT8 arMsg[256];
      OTX_INT32 i;
      OTX_INT32 nMsgLen = 256;
      OTX_HANDLE hHandle;

      hHandle = g_hLogicDevs.hHdlcSender[nCurrentLi];

      for (i = 0; i < nMsgLen; i++) {
        arMsg[i] = (OTX_UINT8) i;
      }

      DemoLogf("Buffer 10 HDLC frames to be transmitted, then Flush the HDLC Sender on Li #%d ...", nCurrentLi);

      // Send the HDLC frame
      for (i = 0; i < 10; i++) {
        result = OtxHdlcSendData(hHandle, arMsg, nMsgLen, DEMO_TASK_REF, 200); DemoCheckRc(hHandle, result);
      }

      DemoLogf("Sleeping...\n");
      OsaThreadSleep(50);

      result = OtxHdlcSenderFlush(hHandle); DemoCheckRc(hHandle, result);



      DemoLogf("Done.\n");
    }
    break;


  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
    // Select current Li
    {
      OTX_INT32 nLi = commandChar - '0';
      
      if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCI && nLi > 2) {
        DemoLogf("Invalid Li# for Thor-2-PCI board (%d)\n", nLi);
        break;
      }
      
      nCurrentLi = nLi;
      DemoLogf("LI#%d Selected ....\n", nCurrentLi);
    }
    break;


#if 0
  case 'f':
    {
      OTX_UINT8 arMsg[64];
      OTX_INT32 i;
      OTX_INT32 nMsgLen = 10;
      OTX_BOOL  bContinue = OTX_TRUE;

      for (i = 0; i < nMsgLen; i++) {
        arMsg[i] = 0xAA;
      }

      DemoLogf("Sending HDLC frames (TS16) from Li #%d ...", nCurrentLi);

      do {  // Send HDLC Frames until user presses 'f' again
        result = OtxHdlcSendData(g_hLogicDevs.hHdlcSender[nCurrentLi], arMsg, nMsgLen, DEMO_TASK_REF, 200);
        commandChar = getch();
        if (commandChar == 'f' || toupper(commandChar) == 'Q') bContinue = OTX_FALSE;

      } while (bContinue);
      DemoLogf("Done.\n");
    }
    break;
#endif

  case 'r':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Li%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'u':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from an 8-bit register in QuadE1T1%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs.hQuadE1T1[nCurrentLi], nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nCurrentLi], result);

      DemoLogf("QuadE1T1%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'R':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from a 16-bit register in the Thor-2-PCI device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg16(g_hPhysDevs.hThorController0, nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      DemoLogf("Brd Address 0x%X = %.4X\n", nAddress, (OTX_UINT16)readVal);
    }
    break;

  case 'w':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, nValue); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Wrote %.2X to Li%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;

  case 'j':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to an 8-bit register in QuadE1T1%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hQuadE1T1[nCurrentLi], nAddress, nValue); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nCurrentLi], result);

      DemoLogf("Wrote %.2X to QuadE1T1%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;

  case 'W':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to a 16-bit register in the Thor-2-PCI device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg16(g_hPhysDevs.hThorController0, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      DemoLogf("Wrote %.4X to Brd Address 0x%X\n", (OTX_UINT16)nValue, nAddress);
    }
    break;

  case 'L':
    {
      char strBuf[12];
      DemoLogf("Enter HDLC Frame Length to be transmitted (%d) ", g_cbHdlcFrame);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_cbHdlcFrame = (OTX_INT32)strtol(strBuf, NULL, 0);
        DemoLogf("HDLC Frame Length is now set to %d\n", g_cbHdlcFrame);
      } else {
        DemoLogf("HDLC Frame Length remains unchanged\n");
      }
    }
    break;

  case 'S':
    {
      char strBuf[12];
      DemoLogf("Enter Time to Sleep (in ms) between transmitted HDLC Frames (%d) ", g_nSleepTimeBetweenHdlcFrames);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_nSleepTimeBetweenHdlcFrames = (OTX_INT32)strtol(strBuf, NULL, 0);
        DemoLogf("Sleep time is now set to %d ms\n", g_nSleepTimeBetweenHdlcFrames);
      } else {
        DemoLogf("Sleep time remains unchanged\n");
      }
    }
    break;

  case '+':
    {
      char strBuf[12];

      DemoLogf("Set the number of consecutive frame to sent when 't' are pressed (%d)\n", g_coHdlcSenderFrames);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_coHdlcSenderFrames = (OTX_INT32) strtol(strBuf, NULL, 0);
        DemoLogf("Consecutive frame count is now %d\n", g_coHdlcSenderFrames);
      } else {
        DemoLogf("Consecutive frame count remains unchanged\n");
      }
    }
    break;



  case 'v':
    {
      OTX_INT32 i;
      g_bHdlcVerbose = !g_bHdlcVerbose;
      DemoLogf("g_bHdlcVerbose is now %s.\n", g_bHdlcVerbose ? "TRUE": "FALSE");
      for (i = 0; i < 32; i++) {
        if (g_coHdlcRxFrames[i]) {
          DemoLogf("g_coHdlcRxFrames[%d] = %d (reset to 0) \n", i, g_coHdlcRxFrames[i]);
          g_coHdlcRxFrames[i] = 0;
          g_coHdlcBadRxFrames[i] = 0;
        }
      }
    }
    break;

  case 'V':
    {
      OTX_INT32 i;
      for (i = 0; i < 32; i++) {
        if (g_coHdlcRxFrames[i]) {
          DemoLogf("Tot[%.2d]=%.6d (Bad[%.2d]=%.6d)  ", i, g_coHdlcRxFrames[i], i, g_coHdlcBadRxFrames[i]);
          if (i && i % 2) {
            DemoLogf("\n");
          }
        }
      }
      DemoLogf("\n");
    }
    break;

  case 'c':
    {
      OtxBrdClockSourceE  eClkSource;
      OTX_INT32 c;

      DemoLogf("Set Clock Source:\n");
      if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) {
        printf("0 = OTX_CLOCK_SOURCE_LOCAL_0  - LIs synchronized to Li0 receive span (QuadE1T1#0)\n");
        printf("1 = OTX_CLOCK_SOURCE_LOCAL_1  - LIs synchronized to Li1 receive span (QuadE1T1#0)\n");
        printf("2 = OTX_CLOCK_SOURCE_LOCAL_2  - LIs synchronized to Li2 receive span (QuadE1T1#0)\n");
        printf("3 = OTX_CLOCK_SOURCE_LOCAL_3  - LIs synchronized to Li3 receive span (QuadE1T1#0)\n");
        printf("4 = OTX_CLOCK_SOURCE_LOCAL_4  - LIs synchronized to Li4 receive span (QuadE1T1#1)\n");
        printf("5 = OTX_CLOCK_SOURCE_LOCAL_5  - LIs synchronized to Li5 receive span (QuadE1T1#1)\n");
        printf("6 = OTX_CLOCK_SOURCE_LOCAL_6  - LIs synchronized to Li6 receive span (QuadE1T1#1)\n");
        printf("7 = OTX_CLOCK_SOURCE_LOCAL_7  - LIs synchronized to Li7 receive span (QuadE1T1#1)\n");
        printf("i = OTX_CLOCK_SOURCE_INTERNAL - LIs synchronized to on-board oscillator\n");

        c = getch();
        if (c == 'i') {
          eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        } else {
          switch (c - '0') {
          case 0:
          case 1:
          case 2:
          case 3:
          case 4:
          case 5:
          case 6:
          case 7:
            eClkSource = (OtxBrdClockSourceE) ( OTX_CLOCK_SOURCE_LOCAL_0 + (c - '0') ); 
            break;
          default:
            printf("Invalid entry. Setting Clock Source to OTX_CLOCK_SOURCE_INTERNAL\n");
            eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
            break;
          }
        }
      } else {
        DemoLogf("0 = OTX_CLOCK_SOURCE_LOCAL_0  - LIs synchronized to Li0 receive span\n");
        DemoLogf("1 = OTX_CLOCK_SOURCE_LOCAL_1  - LIs synchronized to Li1 receive span\n");
        DemoLogf("i = OTX_CLOCK_SOURCE_INTERNAL - LIs synchronized to on-board oscillator\n");

        c = getch();
        if (c == 'i') {
          eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        } else {
          switch (c - '0') {
          case 0:
          case 1:
            eClkSource = (OtxBrdClockSourceE) (OTX_CLOCK_SOURCE_LOCAL_0 + (c - '0') );
            break;
          default:
            DemoLogf("Invalid entry. Setting Clock Source to OTX_CLOCK_SOURCE_INTERNAL\n");
            eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
            break;
          }
        }

      }

      result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, eClkSource); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      DemoLogf("Done.\n");
    }
    break;
  



  case 'O':
    {
      char szBuf[12];
      OTX_INT32 nCmd;
      
      DemoLogf("Driver Debug Command. Which Cmd?\n");
      DemoLogf("0: Dump QuadFalc register\n");
      OsaConsoleGetStr(szBuf);
      nCmd = (OTX_INT32) strtol(szBuf, NULL, 0);
      
      DemoLogf("Issuing Driver Debug Command %d on QuadE1T1%d \n", nCmd, nCurrentLi);
      OtxDrvDbgDump(g_hPhysDevs.hQuadE1T1[nCurrentLi], nCmd);
    }
    break;




#if 1 // Just for debugging
  case '{':
    {
      char strBuf[12];
      static OTX_UINT8 nTssRegister = 0;
      OTX_UINT32 nReadVal;

      DemoLogf("Read from TSS control register. Which control register? [0x%.2X] ", nTssRegister);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nTssRegister = (OTX_UINT8)strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hTss0, 0x01, nTssRegister); DemoCheckRc(g_hPhysDevs.hTss0, result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hTss0, 0x02, 0x00); DemoCheckRc(g_hPhysDevs.hTss0, result);

      result = OtxDrvReadReg8(g_hPhysDevs.hTss0,  0x03, &nReadVal);  DemoCheckRc(g_hPhysDevs.hTss0, result);

      DemoLogf("Read 0x%.2X from TSS Control Register 0x%.2X\n", (OTX_UINT8)nReadVal, nTssRegister);
      DemoLogf("Done.\n");
    }
    break;

  case '}':
    {
      char strBuf[12];
      static OTX_UINT8 nTssRegister = 0;
      static OTX_UINT8 nWriteVal = 0;

      DemoLogf("Write to TSS control register. Which control register? [0x%.2X] ", nTssRegister);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nTssRegister = (OTX_UINT8)strtoul(strBuf, NULL, 0);
      }

      DemoLogf("Which value? [0x%.2X] ", nWriteVal);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nWriteVal = (OTX_UINT8)strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hTss0, 0x01, nTssRegister); DemoCheckRc(g_hPhysDevs.hTss0, result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hTss0, 0x02, 0x00); DemoCheckRc(g_hPhysDevs.hTss0, result);

      result = OtxDrvWriteReg8(g_hPhysDevs.hTss0,  0x03, nWriteVal);  DemoCheckRc(g_hPhysDevs.hTss0, result);

      DemoLogf("Wrote 0x%.2X to TSS Control Register 0x%.2X\n", nWriteVal, nTssRegister);
      DemoLogf("Done.\n");
    }
    break;

  case 'i':
    {
      DemoLogf("Enable TSS... \n");
      result = OtxDrvEnable(g_hPhysDevs.hTss0, OTX_TASK_SYNC,0); DemoCheckRc(g_hPhysDevs.hTss0, result);
      DemoLogf("Done.\n");
    }
    break;

  case 'I':
    {
      DemoLogf("Disable TSS... \n");
      result = OtxDrvDisable(g_hPhysDevs.hTss0, OTX_TASK_SYNC,0); DemoCheckRc(g_hPhysDevs.hTss0, result);
      DemoLogf("Done.\n");
    }
    break;

  case '_':
    {
      DemoLogf("Reseting TSS... \n");
      result = OtxDrvReset(g_hPhysDevs.hTss0); DemoCheckRc(g_hPhysDevs.hTss0, result);
      DemoLogf("Done.\n");
    }
    break;
#endif

#if 1  // Just for debugging
  case '*':
    {
      OTX_UINT32 nReadVal;
      OTX_INT32 nLboVal;

      DemoLogf("Set LBO of Li #%d (T1 long haul mode only)\n", nCurrentLi);
      DemoLogf("0 = 0 dB\n");
      DemoLogf("1 = -7.5  dB\n");
      DemoLogf("2 = -15   dB\n");
      DemoLogf("3 = -22.5 dB\n");

      nLboVal = getch() - '0';
      switch (nLboVal) {
      case 0:  // 0 dB
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x26, 0x01); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x25, 0x2B); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x24, 0x7D); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        break;
      case 1: // -7.5dB
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x26, 0x20); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x25, 0x02); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x24, 0x11); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        break;
      case 2: // -15dB
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x26, 0x20); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x25, 0x01); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x24, 0x8E); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        break;
      case 3: // -22.5dB
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x26, 0x20); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x25, 0x01); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x24, 0x09); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        break;
      default:
        DemoLogf("Invalid entry.\n");
        break;
      }

      switch (nLboVal) {
      case 0:  // 0 dB
      case 1: // -7.5dB
      case 2: // -15dB
      case 3: // -22.5dB
        result = OtxDrvReadReg8(g_hPhysDevs.hLi[nCurrentLi], 0x38, &nReadVal); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x38, (nReadVal & 0x3F) | (nLboVal << 6)); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        break;
      default:
        break;
      }

      DemoLogf("Done.\n");
    }
    break;

  case '!':
    {
      OTX_UINT32 nPulseShapeLevel0;
      OTX_UINT32 nPulseShapeLevel1;
      OTX_UINT32 nPulseShapeLevel2;
      OTX_UINT32 nPulseShapeLevel3;
      char strBuf[12];

      DemoLogf("Customizing Transmit Pulse Shape Level of Li #%d \n", nCurrentLi);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_TX_PULSE_MASK_SUB_PULSE_0, &nPulseShapeLevel0); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_TX_PULSE_MASK_SUB_PULSE_1, &nPulseShapeLevel1); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_TX_PULSE_MASK_SUB_PULSE_2, &nPulseShapeLevel2); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_TX_PULSE_MASK_SUB_PULSE_3, &nPulseShapeLevel3); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      
      DemoLogf("nPulseShapeLevel0 [%d]. Enter new value (0-31): ", nPulseShapeLevel0);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nPulseShapeLevel0 = strtoul(strBuf, NULL, 0);
      }
      
      DemoLogf("nPulseShapeLevel1 [%d]. Enter new value (0-31): ", nPulseShapeLevel1);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nPulseShapeLevel1 = strtoul(strBuf, NULL, 0);
      }
      
      DemoLogf("nPulseShapeLevel2 [%d]. Enter new value (0-31): ", nPulseShapeLevel2);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nPulseShapeLevel2 = strtoul(strBuf, NULL, 0);
      }
      
      DemoLogf("nPulseShapeLevel3 [%d]. Enter new value (0-31): ", nPulseShapeLevel3);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nPulseShapeLevel3 = strtoul(strBuf, NULL, 0);
      }
      
      DemoLogf("Writing Sub Pulse Mask %d,%d,%d,%d \n", nPulseShapeLevel0, nPulseShapeLevel1, nPulseShapeLevel2, nPulseShapeLevel3);

      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_TX_PULSE_MASK_SUB_PULSE_0, nPulseShapeLevel0); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_TX_PULSE_MASK_SUB_PULSE_1, nPulseShapeLevel1); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_TX_PULSE_MASK_SUB_PULSE_2, nPulseShapeLevel2); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_TX_PULSE_MASK_SUB_PULSE_3, nPulseShapeLevel3); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Done.\n");
    }
    break;

  case '&':
    {
      OTX_UINT32 readVal;
      OTX_INT32 nLi;
      char *pszEqStatus;
      
      for (nLi = 0; nLi < 2; nLi++) {
        result = OtxDrvReadReg8(g_hPhysDevs.hLi[nLi], 115, &readVal); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        switch ((readVal & 0xC0) >> 6) {
        case 0x00:
          pszEqStatus = "not valid, still adapting";
          break;
        case 0x01:
          pszEqStatus = "valid";
          break;
        case 0x02:
          pszEqStatus = "not valid";
          break;
        case 0x03:
          pszEqStatus = "valid but high noise floor";
          break;
        default:
          pszEqStatus = "unknown";
          break;
        }
        DemoLogf("Receive Equalizer of Li#%d: status=%s attenuation=%2.1fdB (0x%.2X)\n", nLi, pszEqStatus, (readVal & 0x1F)*1.4, readVal );
      }
      
    }
    break; 
    
  case '^':
    {
      char strBuf[12];
      OTX_UINT32 nThresholdInMilliVolts;

      DemoLogf("Set Receive Input Threshold for Li%d \n", nCurrentLi);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nThresholdInMilliVolts); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Threshold (%d mV): ", nThresholdInMilliVolts);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nThresholdInMilliVolts = strtoul(strBuf, NULL, 0);

        result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, nThresholdInMilliVolts); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

        result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nThresholdInMilliVolts); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        DemoLogf("Receive Input Threshold is changed to %d mV.\n", nThresholdInMilliVolts);
      } else {
        DemoLogf("The Receive Input Threshold remains unchanged.\n");
      }

    }
    break;

  case 'k':
    {
      OtxBoardDeviceAttributesE nAttrib;

      DemoLogf("Turn Amp On for Li%d (HMA-1079-2-1.0 dongle only)\n", nCurrentLi);

      if (nCurrentLi == 0) {
        nAttrib = OTX_ATTR_BRD_PCMCIA_AMP_DONGLE_0;
      } else if (nCurrentLi == 1) {
        nAttrib = OTX_ATTR_BRD_PCMCIA_AMP_DONGLE_1;
      } else {
        DemoLogf("Invalid Li (%d) for this action\n", nCurrentLi);
        break;
      }

      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hThorController0, nAttrib, OTX_TRUE); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      DemoLogf("Done\n");
    }
    break;
  case 'K':
    {
      OtxBoardDeviceAttributesE nAttrib;

      DemoLogf("Turn Amp Off for Li%d (HMA-1079-2-1.0 dongle only)\n", nCurrentLi);

      if (nCurrentLi == 0) {
        nAttrib = OTX_ATTR_BRD_PCMCIA_AMP_DONGLE_0;
      } else if (nCurrentLi == 1) {
        nAttrib = OTX_ATTR_BRD_PCMCIA_AMP_DONGLE_1;
      } else {
        DemoLogf("Invalid Li (%d) for this action\n", nCurrentLi);
        break;
      }

      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hThorController0, nAttrib, OTX_FALSE); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      DemoLogf("Done\n");
    }
    break;



  case '@':
    {

      DemoLogf("Set idle pattern to 0x01 (to facilitate pulse shape testing) for Li #%d \n", nCurrentLi);

      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x29, 0x01); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x30, 0xFF); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x31, 0xFF); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x32, 0xFF); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x33, 0xFF); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Done.\n");
    }
    break;

#endif




  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}



/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->logFileName       = DEMO_DEFAULT_LOG_FILE;
  envP->nBoardType        = OTX_DEVICE_THOR_PCI;
  envP->coLi              = 2;
  envP->law               = OTX_DATA_FORMAT_COMPAND8_U_LAW;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->eT1E1Mode         = OTX_T1E1_MODE_T1;
  envP->bSs7FisuFiltering = OTX_FALSE;
  envP->bLoopHandsetsInternally = OTX_FALSE;
  envP->pszDongleType     = NULL;
  envP->nLineLength       = 10;

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    OsaStdStrUpr(argv[i]);
    if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
      if (*(argv[i]+1) == 'L') {   // Log file
        envP->logFileName = argv[i]+2;
      } else if (*(argv[i]+1) == 'B') {   // Board number
        envP->boardNo = strtoul(argv[i]+2, NULL, 0);
      } else if (*(argv[i]+1) == 'I') {   // Internal handset loop
        envP->bLoopHandsetsInternally = OTX_TRUE;
      } else if (*(argv[i]+1) == 'T') {   // Board Type
        if ( *(argv[i]+2) == '2') {
          envP->nBoardType = OTX_DEVICE_THOR_PCI;
          envP->coLi = 2;
        } else if ( *(argv[i]+2) == '8') {
          envP->nBoardType  = OTX_DEVICE_THOR_8_PCI;
          envP->coLi = 8;
        } else if ( *(argv[i]+2) == 'M') {
          envP->nBoardType  = OTX_DEVICE_THOR_PCMCIA;
          envP->coLi = 2;
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          exit(1);
        }
      } else if (*(argv[i]+1) == 'D') {   // Dongle Type (for Thor-2-PCMCIA)
        if ( *(argv[i]+2) == '1') {
          envP->pszDongleType     = "HMA-1021-1-1.0";     // Thor-2-PCMCIA Standard Dongle (HMA-1021-1-1.0)
        } else if ( *(argv[i]+2) == '2') {
          envP->pszDongleType     = "HMA-1074-1-1.0-P1";  // Thor-2-PCMCIA Monitor Dongle  (HMA-1074-1-1.0-P1)
        } else if ( *(argv[i]+2) == '3') {
          envP->pszDongleType     = "HMA-1076-2-1.0-P1";  // Thor-2-PCMCIA-Plus Generator Dongle (HMA-1076-2-1.0-P1)
        } else if ( *(argv[i]+2) == '4') {
          envP->pszDongleType     = "HMA-1079-2-1.0";     // Thor-2-PCMCIA-Plus Monitor Dongle switchable (HMA-1079-2-1.0)
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          exit(1);
        }
      } else if (*(argv[i]+1) == 'C') {   // Companding Law
        if (*(argv[i]+2) == 'A') {   // A-law
          envP->law = OTX_DATA_FORMAT_COMPAND8_A_LAW;
        } else if (*(argv[i]+2) == 'U') {   // U-law
          envP->law = OTX_DATA_FORMAT_COMPAND8_U_LAW;
        } else {
          DemoLogf("Error: Invalid companding law command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
					exit(1);
        }
      } else if (*(argv[i]+1) == 'M') {   // Li Mode
        if (*(argv[i]+2) == 'E') {   // E1 
          envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
        } else if (*(argv[i]+2) == 'T') {   // T1
          envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
        } else {
          DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
					exit(1);
        }
      } else if (*(argv[i]+1) == 'S') {   // SS7 FISU filtering
        envP->bSs7FisuFiltering = OTX_TRUE;
      } else if (*(argv[i]+1) == 'N') {   // Line length
        envP->nLineLength = (OTX_UINT16)strtoul(argv[i]+2, NULL, 0);
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
				exit(1);
      }
    } else {
      DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
      DemoHelpPanel();
      OsaConsoleExit();
			exit(1);
    }
  }

  if (envP->nBoardType == OTX_DEVICE_THOR_PCMCIA && !envP->pszDongleType) {
    DemoLogf("Error: Dongle type must be supplied for the Thor-2-PCMCIA card\n");
    DemoHelpPanel();
    OsaConsoleExit();
  	exit(1);
  }


}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>    : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[2|8|M]      : Board Type used for the demo, where\n");
  DemoLogf("                    2 = Use Thor-2-PCI Board\n");
  DemoLogf("                    8 = Use Thor-8-PCI Board\n");
  DemoLogf("                    M = Use Thor-2-PCMCIA Board\n");
  DemoLogf(" /L<fileName>   : Log file name\n");
  DemoLogf(" /S             : SS7 FISU filtering active\n");
  DemoLogf(" /CA            : A-law\n");
  DemoLogf(" /CU            : u-law\n");
  DemoLogf(" /ME            : E1 Li mode\n");
  DemoLogf(" /MT            : T1 Li mode\n");
  DemoLogf(" /I             : Internal Handset loop (not via the T1/E1 span)\n");
  DemoLogf(" /D[1|2|3|4]    : Dongle type (only for Thor-2-PCMCIA)\n");
  DemoLogf("                    1 = Thor-2-PCMCIA Standard Dongle (HMA-1021-1-1.0)\n");
  DemoLogf("                    2 = Thor-2-PCMCIA Monitor Dongle  (HMA-1074-1-1.0-P1)\n");
  DemoLogf("                    3 = Thor-2-PCMCIA-Plus Generator Dongle (HMA-1076-2-1.0-P1)\n");
  DemoLogf("                    4 = Thor-2-PCMCIA-Plus Monitor Dongle switchable (HMA-1079-2-1.0)\n");
  DemoLogf(" /N<lineLength> : T1/E1 span line length\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}



/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  DemoLogf("THORDEMO, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("E1T1 Device Example program.\n");
  DemoLogf("For Thor-2-PCI, Thor-8-PCI and Thor-2-PCMCIA\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }

  DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);


  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }

  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Query driver revision
  {
    char szVersionInfo[255];
    result = OtxDrvGetAttributeValueSz(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Driver version %s\n", szVersionInfo);
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnablePhysicalDevices();

  result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  //result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoInitializeLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnableLogicalDevices();


  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
    free(pszCpldBuildNumbers);
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....

  OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300);

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = getch();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
  OsaConsoleExit();

  return 0;
}
